<?php
/**
* Looks for documented and undocumented functions within a block of php code
* @author	Ulf Wendel <ulf@redsys.de>
* @version 0.2
*/
class PhpdocFunctionParser extends PhpdocVariableParser {

	/**
	* Internal structur of a function.
	* @var	array	$emptyFunction
	* @see	functions, analyseFunctionParagraph()
	*/
	var $emptyFunction = array(
	
															"name"					=> "",
															"undoc"					=> true,
															
															"args"					=> array(),
															"params"				=> array()
												);
		
	/**
	* Array of tags that are allowed in front of the function keyword
	* @var	array	$functionTags
	* @see	analyseFunctionParagraph()
	*/
	var $functionTags = array(
															"parameter"	 	=> true,
															"param"				=> true,
															
															"return"			=> true,
															
															"access"			=> true,
															"abstract"		=> true,
															"static"			=> true,
															
															"throws"			=> true,
															
															"see"					=> true,
															"link"				=> true,
															
															"global"			=> true,
															
															"version"			=> true,
															"since"				=> true,
															
															"deprecated"	=> true,
															"deprec"			=> true,
															
															"brother"			=> true,
															"sister"			=> true,
															
															"exclude"		=> true,
															"magic"				=> true,
															
															"author"			=> true,
															"copyright"		=> true,
															
															"todo"				=> true
											);
	
	/**
	* Analyses a function doc comment.
	* @param	array
	* @return array
	*/
	function analyseFunction($para) {
	
		$function = $this->emptyFunction;
		$function["name"] = $para["name"];		

		if (""!=$para["doc"]) {

			$function = $this->analyseTags($this->getTags($para["doc"]), $function, $this->functionTags);
			
			list($msg, $function) = $this->checkParserErrors($function, "function");
			if (""!=$msg) 
				$this->warn->addDocWarning($this->currentFile, "function", $function["name"], $msg, "mismatch");
			
			list($function["sdesc"], $function["desc"]) = $this->getDescription($para["doc"]);
			
			$function["undoc"] = false;
		}

		$function["args"] = $this->getFunctionArgs($para["head"]);			
		$function = $this->checkArgDocs($function);			

#$this->introspection("function", $function);

		return $function;
	} // end func analyseFunction
	
	/**
	* Compares the argument list generated from the function head with the @param tags found.
	*
	* PHPDoc is able to recognize these documentation mistakes:
	* - too few or too many @param tags
	* - name does not match or is missing
	* - type does not match or is missing
	*
	* @param	array							Complete function array
	* @return	array	$function		Function array with 
	*/	
	function checkArgDocs($function) {

		// "args" contains the informations the parser found in the function head.
		// "param" contains the information from the @param tags.
		$num_args 	= count($function["args"]);
		$num_params = count($function["params"]);
		
		// no args? return...
		if (0==$num_args && 0==$num_params) {

			unset($function["args"]);
			return $function;
			
		}
		
		// no args but @param used
		if (0==$num_args && $num_params>0) {
			
			$msg = "Function head shows no parameters, remove all @param tags.";
			$this->warn->addDocWarning($this->currentFile, "function", $function["name"], $msg, "mismatch");
			$function["params"] = array();
			unset($function["args"]);
			return $function;
			
		}

		// compare the informations from the parser with the @param tags
		reset($function["args"]);
		while (list($k, $args)=each($function["args"])) {

#$this->introspection("args", $args);	

			if (isset($function["params"][$k])) {
			
				if ($args["optional"])
					$function["params"][$k]["default"] = $args["default"];

				if (""!=$args["type"] && ""!=$function["params"][$k]["type"] && $args["type"]!=$function["params"][$k]["type"]) {
					
					$type = $args["type"];
					$msg = sprintf("%dth parameter '%s' type does match the the documented type '%s', possible error consider an update to '@param %s %s %s' or '@param %s %s', the variable name is optional.",
													$k+1,
													$args["name"],
													$function["params"][$k]["type"],
													$type,
													$args["name"],
													(isset($function["params"][$k]["desc"])) ? $function["params"][$k]["desc"] : "(description)",
													$type,
													(isset($function["params"][$k]["desc"])) ? $function["params"][$k]["desc"] : "(description)"
												);
					$this->warn->addDocWarning($this->currentFile, "function", $function["name"], $msg, "mismatch");
					$function["params"][$k]["type"] = $type;												
					
				} else if (""!=$function["params"][$k]["type"]) {
				
					$type = $function["params"][$k]["type"];
					
				} else {
				
					$type = "(object objectname|type)";
					
				}
				
				if (""!=$function["params"][$k]["name"] && $args["name"]!=$function["params"][$k]["name"]) {
										
					$msg = sprintf("%dth parameter '%s' does not match the documented name '%s', update the tag to '@param %s %s %s' or '@param %s %s', the variable name is optional.",
													$k+1,
													$args["name"],
													$function["params"][$k]["name"],
													$type,
													$args["name"],
													(isset($function["params"][$k]["desc"])) ? $function["params"][$k]["desc"] : "(description)",
													$type,
													(isset($function["params"][$k]["desc"])) ? $function["params"][$k]["desc"] : "(description)"
												);
					$this->warn->addDocWarning($this->currentFile, "function", $function["name"], $msg, "mismatch");
					$function["params"][$k]["name"] = $args["name"];
																	
				} else if (""==$function["params"][$k]["name"]) {
				
					$function["params"][$k]["name"] = $args["name"];
					
				}
				
			} else {

				$msg = sprintf("%dth parameter '%s' is not documented add '@param %s [description]' to the end of the @param[eter] list.",
													$k+1,
													$args["name"],
													(""==$args["type"]) ? "(object objectname|type)" : $args["type"]
											);

				$function["params"][$k]["name"] 	= $args["name"];
				$function["params"][$k]["undoc"] 	= true;
				
				if (""!=$args["type"])
					$function["params"][$k]["type"] = $args["type"];
					
				if ($args["optional"])
					$function["params"][$k]["default"] = $args["default"];
															
				$this->warn->addDocWarning($this->currentFile, "function", $function["name"], $msg, "missing");
			}
		}
		
		// more @params specified than variables where found in the function head, delete them
		if ($num_params>$k) {
			
			$msg = "The parser found '$num_args' parameter but '$num_params' @param[eter] tags. You should update the @param[eter] list.";
			$this->warn->addDocWarning($this->currentFile, "function", 	$function["name"], $msg, "mismatch");
			for ($i=$k+1;  $i<$num_params; $i++) 
				unset($function["params"][$i]);			
				
		}
		
		unset($function["args"]);
		return $function;
	} // end func checkArgDocs
	
	/**
	* Analyses a function head and returns an array of arguments.
	* @param	string	PHP code to examine.
	* @return	array		Array of arguments: $args[] = array( optional, default, type, name ).
	* @see	getVariableTypeAndValue()
	*/				
	function getFunctionArgs($code) {

		$args = array();
		while (preg_match($this->PHP_COMPLEX["argument"], $code, $regs)) {

			$type 		= "";
			$value 		= "";
			$optional = false;
		
			if (!isset($regs[3])) {
				
				$len_of_value = strlen($regs[1]);
				
			} else if ("=" == $regs[3]) {
		
				$find 	= $regs[1].$regs[2];
				$code 	= substr($code, strpos($code, $find)+strlen($find) );
				
				list ($type, $value, $raw_value) = $this->getVariableTypeAndValue($code);
				$len_of_value = strlen($raw_value);
				$optional = true;
			
			} else {
				
				$len_of_value = strlen($regs[1].$regs[2]);
				
			}
			
			$code = substr($code, $len_of_value);
			$args[] = array(
												"optional" => $optional, 
												"default"  => $value,
												"type"		 => $type,
												"name"		 => $regs[1]
											);
														
		}

		return $args;
	} // end func getFunctionArgs
	
} // end class PhpdocFunctionParser
?>